using UnityEngine;
using System.Collections;
using System.Collections.Generic;

public class Train : MonoBehaviour {

	private AudioSource audioSource;
	public AudioClip trainMove,doorOpen;

    [Tooltip("Number of seconds between waves.")]
    [SerializeField] private float MaxTimeBetweenWaves = 3f;
    [SerializeField] private int dronesPerWave = 50;
    [SerializeField] private float waitTimeAtStation = 1f;
    // for a nice dropdown of directions
    private enum TrainDirections {
        up,
        down,
        left,
        right
    }

    [Space()]
    [Tooltip("Direction the train travels to reach its destination")]
    [SerializeField]private TrainDirections trainDirection = TrainDirections.down;

    [Tooltip("Determines the starting position of this train.")]
    [SerializeField] private float startPositionOffset = 100;

    [Tooltip("Speed the train travels in units/second.")]
	[SerializeField] private float trainSpeed = 0.15f;

    private bool spawning = false;

    private float timeUntilNextWave = 0;

    private Vector3 dropoffPoint;

    private DroneSpawner[] droneSpawners;

    void Start() {
        // Store the dropoff position for later
        dropoffPoint = transform.position;
        droneSpawners = GetComponentsInChildren<DroneSpawner>();

        // Max gives up trying to figure out the correct rotation vector
        if (trainDirection == TrainDirections.down) {
            transform.rotation = Quaternion.Euler(0, 0, 180);
        }

		audioSource = GetComponent<AudioSource>();
    }

	void Update () {
        if (spawning) {
            return;
        }

        // Start a spawn cycle if we've reached the appropriate time
        if ( timeUntilNextWave <= 0 ) {
            StartCoroutine(DoSpawnCycle());
        } else {
            timeUntilNextWave -= Time.deltaTime;
        }

	}



    private IEnumerator DoSpawnCycle() {
        spawning = true;

        // Put the train at start position
        transform.position = transform.position - offsetVector;

		audioSource.PlayOneShot(doorOpen);

        // Lerp the train into the station
        yield return StartCoroutine(LerpTrain(transform.position - offsetVector, dropoffPoint));

        // Spawn the drones and wait until they've all left the train
        foreach (DroneSpawner spawner in droneSpawners) {
            spawner.SpawnWave(dronesPerWave);
        }
        // Wait for the drones to spawn.
        yield return new WaitForSeconds(droneSpawners[0].spawnRate * dronesPerWave);

        // Wait at the station for enemy drones to board.
        yield return new WaitForSeconds(waitTimeAtStation);

        // Leave the station
        yield return StartCoroutine(LerpTrain(dropoffPoint, transform.position + offsetVector));
        spawning = false;
        timeUntilNextWave = waitTimeAtStation + Random.Range(1, MaxTimeBetweenWaves);
        
        yield return null;
    }

    private IEnumerator LerpTrain(Vector3 startPoint, Vector3 endPoint ) {
        transform.position = startPoint;

		audioSource.PlayOneShot(trainMove);

        while ((transform.position - endPoint).magnitude > 0.5f) {
            transform.position = Vector3.Lerp(transform.position, endPoint, trainSpeed);
            yield return new WaitForEndOfFrame();
        }
        transform.position = endPoint;
        yield return null;
    }


    /// <summary>
    /// Converts <see cref="trainDirection"/> into a direction vector.
    /// </summary>
   
    private Vector3 offsetVector { get {
            Vector3 directionVector;
            switch ( trainDirection ) {
                case TrainDirections.up:
                    directionVector = Vector3.up;
                    break;
                case TrainDirections.down:
                    directionVector = Vector3.down;
                    break;
                case TrainDirections.left:
                    directionVector = Vector3.left;
                    break;
                case TrainDirections.right:
                    directionVector = Vector3.right;
                    break;
                default:
                    directionVector = Vector3.one;
                    break;
            }
            return directionVector * startPositionOffset;
        }
    }


    void OnDrawGizmosSelected () {
        Vector3 sp = transform.position - offsetVector;
        Vector3 ep = transform.position + offsetVector;

        Gizmos.color = Color.cyan;
        // Start position to drop-off point
        Gizmos.DrawLine( sp, transform.position);
        Gizmos.DrawSphere(sp, 1f);

        // drop-off point to End Position
        Gizmos.DrawLine(transform.position, ep);
        Gizmos.DrawCube(ep, Vector3.one * 1f);
    }



}
